package cn.iocoder.yudao.module.reward.service.reward;

import cn.hutool.core.collection.CollectionUtil;
import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.framework.mybatis.core.query.LambdaQuery;
import cn.iocoder.yudao.framework.mybatis.core.service.AbstractService;
import cn.iocoder.yudao.framework.mybatis.core.vo.PageVO;
import cn.iocoder.yudao.module.ecw.api.node.NodeApi;
import cn.iocoder.yudao.module.ecw.api.node.dto.NodeRespDTO;
import cn.iocoder.yudao.module.member.api.user.MemberUserApi;
import cn.iocoder.yudao.module.member.api.user.dto.UserRespDTO;
import cn.iocoder.yudao.module.reward.controller.app.reward.vo.AppRewardBackVO;
import cn.iocoder.yudao.module.reward.controller.app.reward.vo.AppRewardListReqVO;
import cn.iocoder.yudao.module.reward.convert.reward.RewardConvert;
import cn.iocoder.yudao.module.reward.dal.dataobject.reward.RewardDO;
import cn.iocoder.yudao.module.reward.dal.mysql.reward.RewardMapper;
import cn.iocoder.yudao.module.reward.enums.QuantitativeRelationSymbolEnum;
import cn.iocoder.yudao.module.reward.enums.RewardStatusEnum;
import cn.iocoder.yudao.module.reward.vo.reward.*;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.RandomStringUtils;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;

import javax.annotation.Resource;
import java.time.Instant;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.reward.enums.ErrorCodeConstants.*;

/**
 * 礼品 Service 实现类
 *
 * @author 系统管理员
 */
@Service
@Validated
public class RewardServiceImpl extends AbstractService<RewardMapper, RewardDO> implements RewardService {

    @Resource
    private RewardMapper rewardMapper;
    @Resource
    private NodeApi nodeApi;
    @Resource
    private MemberUserApi memberUserApi;

    @Override
    public Long create(RewardCreateReqVO createReqVO) {
        //活动开始时间不能比活动结束时间晚，活动结束时间不能比当前时间早
        if (ObjectUtils.allNotNull(createReqVO.getStartTime(), createReqVO.getEndTime())) {
            if (createReqVO.getStartTime().after((createReqVO.getEndTime())) || createReqVO.getEndTime().before(Date.from(Instant.now()))) {
                throw exception(REWARD_START_OR_END_TIME_NOT_ALLOW);
            }
        } else if (ObjectUtils.allNull(createReqVO.getStartTime(), createReqVO.getEndTime())) {
        } else {
            throw exception(REWARD_START_OR_END_TIME_NOT_ALLOW);
        }
        // 创建的礼品状态只能是启用或未启用
        if (createReqVO.getStatus() != RewardStatusEnum.ENABLED.getValue() && createReqVO.getStatus() != RewardStatusEnum.DISABLED.getValue()) {
            throw exception(REWARD_STATUS_NOT_ALLOW_CREATE);
        }
        // 插入
        RewardDO rewardDO = RewardConvert.INSTANCE.convert(createReqVO);
        rewardDO.setCode(generateRewardCode());
        // 剩余数量若没传给默认值0
        if (rewardDO.getQuantityRemain() == null) {
            rewardDO.setQuantityRemain(0);
        }
        //如果插入失败，重新生成code再次插入
        try {
            rewardMapper.insert(rewardDO);
        } catch (Exception e) {
            rewardDO.setCode(generateRewardCode());
            rewardMapper.insert(rewardDO);
        }
        // 返回
        return rewardDO.getId();
    }

    @Override
    public void update(RewardUpdateReqVO updateReqVO) {
        // 校验存在
        RewardDO rewardDO = rewardMapper.selectById(updateReqVO.getId());
        if (rewardDO == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
        RewardDO updateObj = new RewardDO();
        //未启用可编辑所有内容
        if (rewardDO.getStatus() == RewardStatusEnum.DISABLED.getValue()) {
            if (ObjectUtils.allNotNull(updateReqVO.getStartTime(), updateReqVO.getEndTime())) {
                if (updateReqVO.getStartTime().after((updateReqVO.getEndTime())) || updateReqVO.getEndTime().before(Date.from(Instant.now()))) {
                    throw exception(REWARD_START_OR_END_TIME_NOT_ALLOW);
                }
            } else if (ObjectUtils.allNull(updateReqVO.getStartTime(), updateReqVO.getEndTime())) {
            } else {
                throw exception(REWARD_START_OR_END_TIME_NOT_ALLOW);
            }
            updateObj = RewardConvert.INSTANCE.convert(updateReqVO);
        } else if (rewardDO.getStatus() == RewardStatusEnum.ENABLED.getValue()) { //已启用可编辑剩余数量和备注
            updateObj.setQuantityRemain(updateReqVO.getQuantityRemain());
            updateObj.setRemarkEn(updateReqVO.getRemarkEn());
            updateObj.setRemarkFr(updateReqVO.getRemarkFr());
            updateObj.setRemarkZh(updateReqVO.getRemarkZh());
        } else {
            throw exception(REWARD_STATUS_NOT_ALLOW_UPDATE);
        }
        // 更新
        rewardMapper.updateById(updateObj);
    }

    @Override
    public void delete(Long id) {
        RewardDO rewardDO = rewardMapper.selectById(id);
        if (rewardDO == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
        //只有未启用可以删除
        if (rewardDO.getStatus() != RewardStatusEnum.DISABLED.getValue()) {
            throw exception(REWARD_STATUS_NOT_ALLOW_DELETE);
        }
        // 删除
        rewardMapper.deleteById(id);
    }

    private void validateExists(Long id) {
        if (rewardMapper.selectById(id) == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
    }

    @Override
    public RewardDO get(Long id) {
        RewardDO rewardDO = rewardMapper.selectById(id);
        //校验是否过期
        if (rewardDO != null) {
            validateExpire(rewardDO);
        }
        return rewardDO;
    }

    @Override
    public List<RewardDO> getList(Collection<Long> ids) {
        List<RewardDO> rewardDOS = rewardMapper.selectBatchIds(ids);
        //校验是否过期
        rewardDOS.forEach(this::validateExpire);
        return rewardDOS;
    }

    @Override
    public PageResult<RewardDO> getPage(RewardQueryVO query, PageVO page) {

        LambdaQuery<RewardDO> lambdaQuery = new LambdaQuery<>();
        //礼品ID
        lambdaQuery.eqIfPresent(RewardDO::getCode, query.getCode());
        //礼品名称，模糊匹配三种语言
        if (query.getTitle() != null) {
            lambdaQuery.and(wrapper ->
                    wrapper.like(RewardDO::getTitleZh, query.getTitle())
                            .or()
                            .like(RewardDO::getTitleEn, query.getTitle())
                            .or()
                            .like(RewardDO::getTitleFr, query.getTitle())
            );
        }
        //兑换积分
        if (query.getPointsRequireSymbol() != null && query.getPointsRequire() != null) {
            lambdaQuery.gt(query.getPointsRequireSymbol() == QuantitativeRelationSymbolEnum.GT.getValue(), RewardDO::getPointsRequire, query.getPointsRequire())
                    .eq(query.getPointsRequireSymbol() == QuantitativeRelationSymbolEnum.EQ.getValue(), RewardDO::getPointsRequire, query.getPointsRequire())
                    .lt(query.getPointsRequireSymbol() == QuantitativeRelationSymbolEnum.LT.getValue(), RewardDO::getPointsRequire, query.getPointsRequire());
        }
        //兑换网点id，领取方式，状态
        lambdaQuery.eqIfPresent(RewardDO::getNodeId, query.getNodeId())
                .eqIfPresent(RewardDO::getPickMethod, query.getPickMethod())
                .eqIfPresent(RewardDO::getStatus, query.getStatus());
        //备注，模糊匹配三种语言
        if (query.getRemark() != null) {
            lambdaQuery.and(wrapper ->
                    wrapper.like(RewardDO::getRemarkZh, query.getRemark())
                            .or()
                            .like(RewardDO::getRemarkEn, query.getRemark())
                            .or()
                            .like(RewardDO::getRemarkFr, query.getRemark())
            );
        }
        //剩余数量
        if (query.getQuantityRemainSymbol() != null && query.getQuantityRemain() != null) {
            lambdaQuery.gt(query.getQuantityRemainSymbol() == QuantitativeRelationSymbolEnum.GT.getValue(), RewardDO::getQuantityRemain, query.getQuantityRemain())
                    .eq(query.getQuantityRemainSymbol() == QuantitativeRelationSymbolEnum.EQ.getValue(), RewardDO::getQuantityRemain, query.getQuantityRemain())
                    .lt(query.getQuantityRemainSymbol() == QuantitativeRelationSymbolEnum.LT.getValue(), RewardDO::getQuantityRemain, query.getQuantityRemain());
        }
        //创建时间
        lambdaQuery.betweenIfPresent(RewardDO::getCreateTime, query.getBeginCreateTime(), query.getEndCreateTime())
                .orderByDesc(RewardDO::getId);
        PageResult<RewardDO> rewardDOPageResult = rewardMapper.selectPage(page, lambdaQuery);
        //校验是否过期
        rewardDOPageResult.getList().forEach(this::validateExpire);
        return rewardDOPageResult;
    }

    @Override
    public List<RewardDO> getList(RewardQueryVO query) {
        return rewardMapper.selectList(query);
    }

    @Override
    public Long copyReward(Long id) {
        RewardDO reward = rewardMapper.selectById(id);
        if (reward == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
        //重新生成礼品ID
        reward.setCode(generateRewardCode());
        //设置为未启用
        reward.setStatus(RewardStatusEnum.DISABLED.getValue());

        reward.setId(null);
        reward.setCreateTime(null);
        reward.setUpdateTime(null);
        reward.setDeleted(null);
        try {
            rewardMapper.insert(reward);
        } catch (Exception e) {
            reward.setCode(generateRewardCode());
            rewardMapper.insert(reward);
        }
        return reward.getId();
    }

    @Override
    public void delayReward(RewardDelayReqVO delayVO) {
        RewardDO reward = rewardMapper.selectById(delayVO.getId());
        if (reward == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
        //只允许延期已启用状态礼品
        if (reward.getStatus() != RewardStatusEnum.ENABLED.getValue()) {
            throw exception(REWARD_STATUS_NOT_ALLOW_DELAY);
        }
        Instant now = Instant.now();
        //结束时间不能小于当前时间
        if (!delayVO.getEndTime().toInstant().isAfter(now)) {
            throw exception(REWARD_ENDTIME_ERROR);
        }
        rewardMapper.updateById(RewardConvert.INSTANCE.convertDelayReqVO(delayVO));
    }

    @Override
    public void updateStatus(RewardStatusReqVO statusVO) {
        this.validateExists(statusVO.getId());
        RewardDO rewardDO = rewardMapper.selectById(statusVO.getId());
        Integer oldStatus = rewardDO.getStatus();
        Integer newStatus = statusVO.getStatus();
        switch (newStatus) {
            //关闭：未启用情况下，不可进行关闭；关闭后状态为已关闭，不支持兑换
            case 3:
                if (oldStatus == RewardStatusEnum.DISABLED.getValue()) {
                    throw exception(REWARD_STATUS_NOT_ALLOW_ClOSE);
                } else {
                    rewardMapper.updateById(RewardConvert.INSTANCE.convertStatusReqVO(statusVO));
                }
                break;
            //启用：有效期有效且未启用可进行点击启用
            case 1:
                if (oldStatus == RewardStatusEnum.DISABLED.getValue()) {
                    if (rewardDO.getEndTime().toInstant().isBefore(Instant.now())) {
                        throw exception(REWARD_STATUS_NOT_ALLOW_ENABLE);
                    }
                    rewardMapper.updateById(RewardConvert.INSTANCE.convertStatusReqVO(statusVO));
                }
                break;
            default:
                throw exception(REWARD_STATUS_CHANGE_ERROR);
        }
    }

    @Override
    public List<AppRewardBackVO> getListByMemberInfo(AppRewardListReqVO reqVO) {
        LambdaQuery<RewardDO> rewardDOLambdaQuery = new LambdaQuery<>();
        rewardDOLambdaQuery.eq(RewardDO::getStatus, RewardStatusEnum.ENABLED.getValue())
                .gt(RewardDO::getQuantityRemain, 0)
                .le(RewardDO::getStartTime, new Date())
                .ge(RewardDO::getEndTime, new Date());
        // 如果有网点，直接根据网点查
        if (reqVO.getNodeId() != null) {
            rewardDOLambdaQuery.eq(RewardDO::getNodeId, reqVO.getNodeId());
        } else {
            UserRespDTO user = memberUserApi.getUser(reqVO.getMemberId());
            List<NodeRespDTO> nodes = null;
            // 如果用户没有填写国家或城市默认展示中国的礼品
            if (user.getCity() == null || user.getCountry() == null) {
                nodes = nodeApi.getNodesByCity(null,130L);
            } else {
                nodes = nodeApi.getNodesByCity(user.getCity().longValue(),null);
            }
            if (CollectionUtil.isNotEmpty(nodes)) {
                rewardDOLambdaQuery.in(RewardDO::getNodeId, nodes.stream().map(NodeRespDTO::getId).collect(Collectors.toList()));
            }
        }
        List<RewardDO> rewardDOS = rewardMapper.selectList(rewardDOLambdaQuery);
        if (rewardDOS != null && !rewardDOS.isEmpty()) {
            return rewardDOS.stream().map(RewardConvert.INSTANCE::convertAppVO).collect(Collectors.toList());
        }
        return Collections.EMPTY_LIST;
    }

    /**
     * 生成礼品ID
     */
    private String generateRewardCode() {
        return RandomStringUtils.randomAlphanumeric(12).toUpperCase();
    }

    /**
     * 校验礼品是否过期并修改礼品状态
     */
    private void validateExpire(RewardDO rewardDO) {
        //TODO：异步更新礼品状态
        if (rewardDO.getEndTime() != null) {
            if (rewardDO.getStatus() == RewardStatusEnum.DISABLED.getValue() && rewardDO.getEndTime().toInstant().isBefore(Instant.now())) {
                RewardDO expireReward = new RewardDO();
                expireReward.setId(rewardDO.getId());
                expireReward.setStatus(RewardStatusEnum.EXPIRED.getValue());
                rewardMapper.updateById(expireReward);
            }
        }
    }
}
